﻿--[[
    Author: Valana <Team BATTLE> of Mug'thol Alliance-US
]]

local TbArtisan = LibStub("AceAddon-3.0"):NewAddon("TbArtisan", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("TB_Artisan", true)
local TbCore = TbCore

local bband = bit.band
local strlen = strlen
local strmatch = strmatch
local time = time
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local CheckInteractDistance = CheckInteractDistance
local GetNumLootItems = GetNumLootItems
local LootSlot = LootSlot
local ConfirmLootSlot = ConfirmLootSlot
local GetSpellLink = GetSpellLink

local _player = UnitName("player")
local _login = time()
local _me = bit.bor(COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_TYPE_PLAYER)
local _party = bit.bor(COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_TYPE_PLAYER)
local _raid = bit.bor(COMBATLOG_OBJECT_AFFILIATION_RAID, COMBATLOG_OBJECT_TYPE_PLAYER)
local _alchemy = GetSpellInfo(2259)
local _elixirMaster = GetSpellInfo(28677)
local _potionMaster = GetSpellInfo(28675)
local _transmutationMaster = GetSpellInfo(28672)
local _blacksmithing = GetSpellInfo(2018)
local _armorsmith = GetSpellInfo(9788)
local _masterAxesmith = GetSpellInfo(17041)
local _masterHammersmith = GetSpellInfo(17040)
local _masterSwordsmith = GetSpellInfo(17039)
local _cooking = GetSpellInfo(2550)
local _enchanting = GetSpellInfo(7411)
local _engineering = GetSpellInfo(4036)
local _gnomishEngineer = GetSpellInfo(20219)
local _goblinEngineer = GetSpellInfo(20222)
local _firstAid = GetSpellInfo(3273)
local _inscription = GetSpellInfo(45357)
local _jewelcrafting = GetSpellInfo(25229)
local _leatherworking = GetSpellInfo(2108)
local _dragonscaleLeatherworking = GetSpellInfo(10656)
local _elementalLeatherworking = GetSpellInfo(10658)
local _tribalLeatherworking = GetSpellInfo(10660)
local _smelting = GetSpellInfo(2656)
local _tailoring = GetSpellInfo(3908)
local _moonclothTailoring = GetSpellInfo(26798)
local _shadoweaveTailoring = GetSpellInfo(26801)
local _spellfireTailoring = GetSpellInfo(26797)
local _gatherers =
{
    [GetSpellInfo(13262)] = true,       -- Disenchant.
    [GetSpellInfo(31252)] = true,       -- Prospecting.
    [GetSpellInfo(51005)] = true,       -- Milling.
    [GetSpellInfo(8613)] = true,        -- Skinning.
    [GetSpellInfo(7620)] = true,        -- Fishing.
    [GetSpellInfo(2575)] = true,        -- Mining.
    [GetSpellInfo(2366)] = true,        -- Herb Gathering.
    [GetSpellInfo(4036)] = true,        -- Engineering.
    [GetSpellInfo(30427)] = true,       -- Extract Gas.
}

local _possibleSpecializations =
{
    [_alchemy] = { [_elixirMaster] = true, [_potionMaster] = true, [_transmutationMaster] = true },
    [_blacksmithing] = { [_armorsmith] = true, [_masterAxesmith] = true, [_masterHammersmith] = true, [_masterSwordsmith] = true },
    [_engineering] = { [_gnomishEngineer] = true, [_goblinEngineer] = true },
    [_leatherworking] = { [_dragonscaleLeatherworking] = true, [_elementalLeatherworking] = true, [_tribalLeatherworking] = true },
    [_tailoring] = { [_moonclothTailoring] = true, [_shadoweaveTailoring] = true, [_spellfireTailoring] = true },
}

local _currentSpecializations =
{
    [_alchemy] = _alchemy,
    [_blacksmithing] = _blacksmithing,
    [_cooking] = _cooking,
    [_enchanting] = _enchanting,
    [_engineering] = _engineering,
    [_firstAid] = _firstAid,
    [_inscription] = _inscription,
    [_jewelcrafting] = _jewelcrafting,
    [_leatherworking] = _leatherworking,
    [_smelting] = _smelting,
    [_tailoring] = _tailoring,
}


function TbArtisan:OnInitialize()
    --
    -- Load our settings.
    --
    self.db = LibStub("AceDB-3.0"):New("TbArtisanDb",
    {
        global =
        {
			announceFeasts = true,
			thankChefs = true,
		},
        factionrealm =
        {
            [_alchemy] = { },
            [_blacksmithing] = { },
            [_cooking] = { },
            [_enchanting] = { },
            [_engineering] = { },
            [_firstAid] = { },
            [_inscription] = { },
            [_jewelcrafting] = { },
            [_leatherworking] = { },
            [_smelting] = { },
            [_tailoring] = { },
        },
        char =
        {
            includeAlts =
            {
                [_alchemy] = true,
                [_blacksmithing] = true,
                [_cooking] = true,
                [_enchanting] = true,
                [_engineering] = true,
                [_firstAid] = false,
                [_inscription] = true,
                [_jewelcrafting] = true,
                [_leatherworking] = true,
                [_smelting] = false,
                [_tailoring] = true,
            },
        },
    }, "Default")
    self.db:SetProfile("Default")

    --
    -- Create our options menu and add it to the Blizzard interface menu.
    --
    local _, addon = GetAddOnInfo("TB_Artisan")
    LibStub("AceConfig-3.0"):RegisterOptionsTable(addon,
        {
            type = "group",
            name = addon,
            desc = GetAddOnMetadata("TB_Artisan", L["DESCRIPTION_METADATA_FIELD"]),
            handler = TbArtisan,
            childGroups  = "tab",
            args =
            {
                subHeader =
                {
                    order = 0,
                    type = "description",
                    name = L["MENU_SUBHEADER"],
                },
                feasts =
                {
                    order = 1,
                    type = "group",
                    name = L["MENU_FEASTS"],
                    desc = L["MENU_FEASTS_DESCRIPTION"],
                    args =
                    {
                        feastsSubHeader =
                        {
                            order = 0,
                            type = "description",
                            name = L["MENU_FEASTS_SUBHEADER"],
                        },
                        announceFeasts =
                        {
                            order = 10,
                            type = "toggle",
                            name = L["MENU_ANNOUNCE_FEASTS"],
                            desc = L["MENU_ANNOUNCE_FEASTS_DESCRIPTION"],
                            get = "IsAnnouncingFeasts",
                            set = "ToggleAnnounceFeasts",
                        },
                        thankChefs =
                        {
                            order = 20,
                            type = "toggle",
                            name = L["MENU_THANK_CHEFS"],
                            desc = L["MENU_THANK_CHEFS_DESCRIPTION"],
                            get = "IsThankingChefs",
                            set = "ToggleThankChefs",
                        },
                    }
                },
                professionLinks =
                {
                    order = 100,
                    type = "group",
                    name = L["MENU_PROFESSION_LINKS"],
                    desc = L["MENU_PROFESSION_LINKS_DESCRIPTION"],
                    args =
                    {
                        professionLinksSubHeader =
                        {
                            order = 0,
                            type = "description",
                            name = L["MENU_PROFESSION_LINKS_SUBHEADER"],
                        },
                        alchemy =
                        {
                            order = 10,
                            type = "toggle",
                            name = _alchemy,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _alchemy),
                            get = function() return self.db.char.includeAlts[_alchemy] end,
                            set = function() self.db.char.includeAlts[_alchemy] = not self.db.char.includeAlts[_alchemy] end,
                        },
                        blacksmithing =
                        {
                            order = 10,
                            type = "toggle",
                            name = _blacksmithing,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _blacksmithing),
                            get = function() return self.db.char.includeAlts[_blacksmithing] end,
                            set = function() self.db.char.includeAlts[_blacksmithing] = not self.db.char.includeAlts[_blacksmithing] end,
                        },
                        cooking =
                        {
                            order = 10,
                            type = "toggle",
                            name = _cooking,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _cooking),
                            get = function() return self.db.char.includeAlts[_cooking] end,
                            set = function() self.db.char.includeAlts[_cooking] = not self.db.char.includeAlts[_cooking] end,
                        },
                        enchanting =
                        {
                            order = 10,
                            type = "toggle",
                            name = _enchanting,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _enchanting),
                            get = function() return self.db.char.includeAlts[_enchanting] end,
                            set = function() self.db.char.includeAlts[_enchanting] = not self.db.char.includeAlts[_enchanting] end,
                        },
                        engineering =
                        {
                            order = 10,
                            type = "toggle",
                            name = _engineering,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _engineering),
                            get = function() return self.db.char.includeAlts[_engineering] end,
                            set = function() self.db.char.includeAlts[_engineering] = not self.db.char.includeAlts[_engineering] end,
                        },
                        firstAid =
                        {
                            order = 10,
                            type = "toggle",
                            name = _firstAid,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _firstAid),
                            get = function() return self.db.char.includeAlts[_firstAid] end,
                            set = function() self.db.char.includeAlts[_firstAid] = not self.db.char.includeAlts[_firstAid] end,
                        },
                        inscription =
                        {
                            order = 10,
                            type = "toggle",
                            name = _inscription,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _inscription),
                            get = function() return self.db.char.includeAlts[_inscription] end,
                            set = function() self.db.char.includeAlts[_inscription] = not self.db.char.includeAlts[_inscription] end,
                        },
                        jewelcrafting =
                        {
                            order = 10,
                            type = "toggle",
                            name = _jewelcrafting,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _jewelcrafting),
                            get = function() return self.db.char.includeAlts[_jewelcrafting] end,
                            set = function() self.db.char.includeAlts[_jewelcrafting] = not self.db.char.includeAlts[_jewelcrafting] end,
                        },
                        leatherworking =
                        {
                            order = 10,
                            type = "toggle",
                            name = _leatherworking,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _leatherworking),
                            get = function() return self.db.char.includeAlts[_leatherworking] end,
                            set = function() self.db.char.includeAlts[_leatherworking] = not self.db.char.includeAlts[_leatherworking] end,
                        },
                        smelting =
                        {
                            order = 10,
                            type = "toggle",
                            name = _smelting,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _smelting),
                            get = function() return self.db.char.includeAlts[_smelting] end,
                            set = function() self.db.char.includeAlts[_smelting] = not self.db.char.includeAlts[_smelting] end,
                        },
                        tailoring =
                        {
                            order = 10,
                            type = "toggle",
                            name = _tailoring,
                            desc = format(L["MENU_INCLUDE_ALT_PROFESSION_LINKS"], _tailoring),
                            get = function() return self.db.char.includeAlts[_tailoring] end,
                            set = function() self.db.char.includeAlts[_tailoring] = not self.db.char.includeAlts[_tailoring] end,
                        },
                    },
                },
            },
        })
    LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addon)

    --
    -- Disable the old TradeSkills add-ons.
    --
    DisableAddOn("TB_TradeSkills")
    DisableAddOn("TB_TradeSkillApprentice")
end

function TbArtisan:OnEnable()
    --
    -- Apply all our settings.
    --
    self:ApplyAllSettings()

    --
    -- Hook the errors frame to hide looting error messages caused by the player auto-looting.
    --
    self:RawHook(UIErrorsFrame, "AddMessage", "UIErrorsFrame_AddMessage", true)

    --
    -- Listen to successful spellcasts so we know when we've finished breaking down an item.
    --
    self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")

    --
    -- Listen to whispers to detect tradeskills queries.
    --
    self:RegisterEvent("CHAT_MSG_WHISPER")

    --
    -- Watch tradeskill updates so we may remember what our characters are capable of.
    --
    self:RegisterEvent("TRADE_SKILL_UPDATE")

    --
    -- Go through the player's spell book and try to find all of his or her specializations.
    --
    local GetSpellName = GetSpellName
    for i = 1, MAX_SPELLS do
        --
        -- Check the current spell against each possible specialization for any given profession with specializations.
        --
        local spell = GetSpellName(i, BOOKTYPE_SPELL)
        if (not spell) then break end
        local foundSpecialization
        for profession, specializations in pairs(_possibleSpecializations) do
            if (specializations[spell]) then
                --
                -- We've found a specialization. Remember which profession it was for.
                --
                _currentSpecializations[profession] = spell
                foundSpecialization = profession
                break
            end
            if (foundSpecialization) then break end
        end

        --
        -- Prevent this profession's specializations from being searched again.
        --
        if (foundSpecialization) then _possibleSpecializations[foundSpecialization] = nil end
    end
end

--
-- Filter out "Can't loot that item now."
--
function TbArtisan:UIErrorsFrame_AddMessage(frame, message, ...)
    if (message ~= ERR_LOOT_CANT_LOOT_THAT_NOW) then return self.hooks[frame].AddMessage(frame, message, ...) end
end


-------------------------------------------------------------------------------
-- Properties.
-------------------------------------------------------------------------------
function TbArtisan:IsAnnouncingFeasts()
    return self.db.global.announceFeasts
end

function TbArtisan:ToggleAnnounceFeasts()
    self.db.global.announceFeasts = not TbArtisan:IsAnnouncingFeasts()
    self:ApplyAllSettings()
end

function TbArtisan:IsThankingChefs()
    return self.db.global.thankChefs
end

function TbArtisan:ToggleThankChefs()
    self.db.global.thankChefs = not TbArtisan:IsThankingChefs()
    self:ApplyAllSettings()
end


-------------------------------------------------------------------------------
-- Apply the current settings.
-------------------------------------------------------------------------------
function TbArtisan:ApplyAllSettings()
    if (self:IsAnnouncingFeasts() or self:IsThankingChefs()) then
        self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    else
        self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    end
end

-------------------------------------------------------------------------------
-- Automatically loot materials produced by a profession.
-------------------------------------------------------------------------------
function TbArtisan:COMBAT_LOG_EVENT_UNFILTERED(e, timestamp, event, sourceGuid, sourceName, sourceFlags, destinationGuid, destinationName, destinationFlags, spellId, spellName, ...)
    --
    -- We don't care about anything else except Great Feasts and Fish Feasts prepared while we are in a group.
    --
    if ((spellId ~= 57301 and spellId ~= 57426) or not TbCore:IsInGroup() or event ~= "SPELL_CREATE") then return end

    if (self:IsAnnouncingFeasts() and bband(sourceFlags, _me) == _me) then
        --
        -- Determine what kind of feast we prepared.
        --
        local announcement = L["ANNOUNCEMENT_PIG"]
        if (spellId == 57426) then
            announcement = L["ANNOUNCEMENT_FISH"]
        end

        --
        -- Tell our group to stuff themselves.
        --
        if (TbCore:CanUseRaidWarning()) then
            SendChatMessage(announcement, "RAID_WARNING")
        else
            SendChatMessage(announcement, "RAID")
        end
    elseif (self:IsThankingChefs() and not UnitIsDeadOrGhost("player") and (bband(sourceFlags, _party) == _party or bband(sourceFlags, _raid)) and self:IsThankingChefs() and CheckInteractDistance(sourceName, 4) and not TbCore:IsIgnored(sourceName)) then
        SendChatMessage(format(L["THANK_CHEF"], sourceName, spellName), "EMOTE")
    end
end

function TbArtisan:UNIT_SPELLCAST_SUCCEEDED(e, source, spell, rank, ...)
    if (_gatherers[spell] and source == "player") then
        self:RegisterEvent("LOOT_OPENED")
    end
end

function TbArtisan:LOOT_OPENED(e, isAutoLooting)
    --
    -- Automatically loot all materials, unless there is only one item and we are auto-looting (to avoid repeated looting sounds).
    --
    self:UnregisterEvent("LOOT_OPENED")
    if (isAutoLooting == 0 or GetNumLootItems() > 1) then
		self.LootRemainingItems()
	end
end

TbArtisan.LootRemainingItems = function(previous)
    local num = GetNumLootItems()
    if (num == 0) then return end

    --
    -- If there are items to loot, and this is the first call or the server HAS registered our previous loot request, then loot the last stack.
    --
    if (not previous or previous > num) then
        LootSlot(num)
    end

    --
    -- Try again in 100 milliseconds.
    --
    TbArtisan:ScheduleTimer(TbArtisan.LootRemainingItems, .1, num)
end


-------------------------------------------------------------------------------
-- Respond to tradeskill queries via whisper.
-------------------------------------------------------------------------------
function TbArtisan:CHAT_MSG_WHISPER(e, message, sender, ...)
    if (TbCore:IsIgnored(sender)) then return end

    --
    -- First determine if this is even a tradeskill link query.
    --
    local profession, spell = TbArtisan:GetSpellName(TbCore:StrLowerTrim(message))
    if (not profession) then return end

    --
    -- Get the tradeskill link and either respond with it, or notify the sender that we don't have the requested tradeskill. However,
    -- if we have been logged in for less than 30 seconds, fall back on our cache if profession data is not yet available from the
    -- server. Then update our cache with the outcome.
    --
    local _, link = GetSpellLink(spell)
    if (not link and time() - _login < 30) then
		link = self.db.factionrealm[profession][_player]
    end
    self.db.factionrealm[profession][_player] = link

    if (not link) then
        --
        -- Notify the sender that we do not know this profession. Optionally, advise him or her of what our alts can do, if applicable.
        --
        SendChatMessage(L["TRADESKILL_NOT_AVAILABLE"], "WHISPER", nil, sender)
        if (self.db.char.includeAlts[profession] and TbCore:CountDictionaryEntries(self.db.factionrealm[profession]) > 0) then
            SendChatMessage(L["TRADESKILL_ALTS_ONLY"], "WHISPER", nil, sender)
            for alt, link in pairs(self.db.factionrealm[profession]) do
                SendChatMessage("<TB Artisan> " .. alt .. " - " .. link, "WHISPER", nil, sender)
            end
        end
    else
        --
        -- Reply with a profession link. Optionally, advise the sender of what our alts can do, if applicable.
        --
        SendChatMessage("<TB Artisan> " .. link, "WHISPER", nil, sender)
        if (self.db.char.includeAlts[profession] and TbCore:CountDictionaryEntries(self.db.factionrealm[profession]) > 1) then
            SendChatMessage(L["TRADESKILL_ALTS_ALSO"], "WHISPER", nil, sender)
            for alt, link in pairs(self.db.factionrealm[profession]) do
                if (not UnitIsUnit(_player, alt)) then
                    SendChatMessage("<TB Artisan> " .. alt .. " - " .. link, "WHISPER", nil, sender)
                end
            end
        end
    end
end

function TbArtisan:GetSpellName(message)
    if (strlen(message) < 2) then return end

    --
    -- If the whispered message starts with "?", then try to match one of the keywords. The query has to be in the form "?jc".
    --
    local match = strmatch(message, "^%p*[%?!]%p*(.-)%p*$")
    if (match) then
        local profession = L["KEYWORDS"][TbCore:GetTextFromHyperlink(match)]
        local spell = _currentSpecializations[profession]
        if (spell) then return profession, spell end
    end

    --
    -- If the whispered message ends with "?", then try to match one of the keywords. The query has to be in the form "jc?".
    --
    match = strmatch(message, "^%p*([^%p]-)%p*%?%p*$")
    if (match) then
        local profession = L["KEYWORDS"][TbCore:GetTextFromHyperlink(match)]
        local spell = _currentSpecializations[profession]
        if (spell) then return profession, spell end
    end

    --
    -- Look for one of the trigger phrases. This loop gets called on every whisper so L["TRIGGERS"] should be kept as light as possible.
    --
    for index, trigger in ipairs(L["TRIGGERS"]) do
        match = strmatch(message, trigger)
        if (match) then
            --
            -- We've matched a trigger phrase. Now see if we can match one of the tradeskill keywords.
            --
            local profession = L["KEYWORDS"][TbCore:GetTextFromHyperlink(match)]
            local spell = _currentSpecializations[profession]
            if (spell) then return profession, spell end
        end
    end
end


-------------------------------------------------------------------------------
-- Update our database with the player's profession links.
-------------------------------------------------------------------------------
function TbArtisan:TRADE_SKILL_UPDATE(e)
    for profession, specialization in pairs(_currentSpecializations) do
        local _, link = GetSpellLink(specialization)
        if (link) then
			self.db.factionrealm[profession][_player] = link
		end
    end
end
